/* 
 See SampleController.m for more details on how this app works.
 
 
 The SampleItem instance has three variables besides the obvious:
 
 AASBasicInfoFile *file;
 This is the file object that represents the file this SampleItem is concerned
 with. The SampleItem will use the file object to keep track of information
 such as the file's path, filetype, if it's a folder, etc.
 
 AASBasicInfoFile *resolvedFile;
 In the case that this item represents a file that is an alias or symbolic link,
 we might need information about the file it points to (such as  when getting
 contents of a symbolic link to a folder).
 
 AASDirectorySubscription *subscription;
 If this item represents a folder, it'll need to get a list of and keep track of
 all of its children. This is where the subscription object comes in.
 
 The most useful thing in this sample is the implementation of the
 directorySubscription:foundChanges: protocol method at the bottom of this file!
 */


#import "SampleItem.h"


@implementation SampleItem

/* Very basic init method */
- (id)initWithFile:(AASBasicInfoFile *)initFile {
	self = [super init];
	if (self != nil) {
		file = [initFile retain];
		isRoot = NO;
	}
	return self;
}


- (void)dealloc {
	if (subscription) {
		// ALWAYS unsubscribe from the file manager or you'll crash
		// AASFileManager has  other unsubscribe methods that do not require us to keep a pointer to the subscription object. However, the subscription is a useful thing to have around (though this example does not really use it) plus this method is the most efficient way to unsubscribe (see AASFileManagerTracking.h)
		[[AASFileManager defaultFileManager] unsubscribe:self forSubscription:subscription];
	}
	
	[resolvedFile release];
	[file release];
	[children release];
	[childrenWithKeys release];
	
	[super dealloc];
}





- (BOOL)isRoot {
	return isRoot;
}

- (void)setRoot:(BOOL)root {
	isRoot = root;
	
	// Load the children immediately for the root folder.  This will make this object subscribe to filesystem checks immediately.  Since the root item has no parent to keep track of it, this is necessary.  Otherwise if the file got moved/renamed/whatever before its children were loaded it wouldn't get updated and our app would be out of sync.
	if (isRoot && !children) {
		[self haveChildren];
	}
}





/* Returning basic information about our file */
- (AASBasicInfoFile *)file {
	return file;
}

- (BOOL)isDirectory {
	return [file isDirectory];
}

- (NSString *)filename {
	return [file displayName];
}

- (NSString *)path {
	return [file path];
}

- (NSURL *)url {
	return [file url];
}

- (NSString *)type {
	if ([file isInvisible]) return @"Invisible";
	if ([file isSymlink]) return @"Symlink";
	if ([file isAlias]) return @"Alias";
	if ([file isApplication]) return @"Application";
	if ([file isPackage]) return @"Package";
	if ([file isDirectory]) return @"Folder";

	return @"File";
}








/* Load children when they're asked for, and not before. Otherwise we'd recusively load everything in the file heirarchy. */
- (id)childAtIndex:(int)index {
	if (!children) {
		[self haveChildren];
	}
	
	return [children objectAtIndex:index];
}
- (int)numberOfChildren {
	if (!children) {
		[self haveChildren];
	}
	
	return [children count];
}

/* The miracle of birth. */
- (void)haveChildren {
	
	// we shouldn't be calling haveChildren if these exist already, but just in case...
	[children release];	
	[childrenWithKeys release];
		
	// Create containers to hold our children. Because there is only one unique file object for each file, we can use the objects as keys in a dictionary for faster lookup in certain cases.
	children = [[NSMutableArray alloc] initWithCapacity:500];
	childrenWithKeys = [[NSMutableDictionary alloc] initWithCapacity:500];
	
	
	
	if ([file isDirectory]) {
		
		// If the "directory" we're looking at is actually a symbolic link or an alias, we'll run into trouble trying to check its contents. So first, make sure we subscribe using the actual directory.
		if (![file isRealFile]) {
			resolvedFile = [AASBasicInfoFile fileForPath:[file resolvedPath]];
		} else {
			resolvedFile = [file retain]; // This way, we can just use resolvedFile for everything
		}
		
		
		// We hold on to the AASDirectorySubscription object because it might be useful (though not in this example), and so that we can unsubscribe more efficiently later. We don't actually have to keep it though.
		subscription = [[AASFileManager defaultFileManager] subscribe:self forFile:resolvedFile];
		
		if (subscription) {		
			// We successfully subscribed, add the files in this directory to our children array
			// If we're using asynchronous loading (threaded), the subscription might return none or only some of the total files at this point. The remainder will be added via notification later.
			[self addChildren:[subscription files]];
		} else {
			// Laziness drives us to just whine about the problem.
			NSLog(@"There was an error trying to subscribe %@ for file object %@", self, file);
		}
		
		// Lets sort these by filename
		[self sortChildren];
	}
}


/* Methods for adding and removing children */

- (void)addChildren:(NSArray *)newChildren {
	int a, count = [newChildren count];
	
	for (a = 0; a < count; a++) {
		[self addChild:[newChildren objectAtIndex:a]];
	}
}

- (void)addChild:(AASBasicInfoFile *)child {
	SampleItem *childItem = [[SampleItem alloc] initWithFile:child];
	
	// We'll add the item to the children array, plus we will add it to the childrenWithKeys dictionary using the file object as the key (there is a 1 : 1 relationship between file objects and the files on disk). The array lets us keep items in order and have quick access to them, while the dictionary will allow us to efficiently find any item associated with a given file (good when we need to remove them later).
	[children addObject:childItem];
	[childrenWithKeys setObject:childItem forKey:child];
	
	[childItem release];
}

- (void)removeChild:(AASBasicInfoFile *)child {
	// Our dictionary comes in handy
	SampleItem *childItem = [childrenWithKeys objectForKey:child];
	
	[children removeObject:childItem];
	[childrenWithKeys removeObjectForKey:child];
}

- (void)removeAllChildren {
	id temp = children;
	children = nil;
	[temp release];
	
	temp = childrenWithKeys;
	childrenWithKeys = nil;
	[temp release];
}




/* Sorting children */

- (void)sortChildren {
	[children sortUsingSelector:@selector(compareFilename:)];
}

- (NSComparisonResult)compareFilename:(SampleItem *)otherItem {
	return [[self filename] caseInsensitiveCompare:[otherItem filename]];
}







/* Our notification methods are used to tell other interested parties (like the SampleController instance) what's happening. */
- (void)sendUpdateNotification {
	[[NSNotificationCenter defaultCenter] postNotificationName:@"SampleItemUpdatedNotification" object:self];
}

- (void)sendRootRemovalNotification {
	[[NSNotificationCenter defaultCenter] postNotificationName:@"SampleItemRootLostNotification" object:self];
}





/* AASDirectorySubscriptionProtocol method - this is called by the file manager when the folder we subscribed to changes. This is probably the most important part of this sample and will give you a fully usable outline implementation of directorySubscription:foundChanges: that you can integrate into your own application, along with detailed instructions on what you'll need to do in each step. */

- (void)directorySubscription:(AASDirectorySubscription *)sub foundChanges:(NSArray *)changes {
    NSDictionary *fileChangeInfo;
    AASBasicInfoFile *child;
    int a, count = [changes count];
	
    // Go through the changes and do what is necessary
    for (a = 0; a < count; a++) {
				
		// Get the dictionary containing information about this change (see AASDirectorySubscription.h)
        fileChangeInfo = [changes objectAtIndex:a];
		
		// Get the file object that represents the affected file out of the dictionary
        child = [fileChangeInfo valueForKey:AASFileKey];
        
		
		// Now, get the event type out of the dictionary and do something about it!
		switch ([(NSNumber *)[fileChangeInfo valueForKey:AASFileEventKey] intValue]) {
				
			case kAASFileAppearedEvent:
			// Either we have a new file added or we've just initiated the checker and are getting our list of files for the first time. Either way, just add the file to our children.
				[self addChild:child];
				break;

				
			case kAASFileMovedEvent:
			// File moved out of this folder or the folder itself moved.
				
					// Check to make sure it wasn't the directory this item represents that was itself moved (the subscription not only tracks files in the subscribed directory, but also the directory itself). When used properly the AASFilesystemTools framework ensures there's only one unique file object per file in the filesystem, so we can efficiently and safely compare the objects by pointer value.
				if (child == file) {
					// Do what needs to be done if the folder we're watching moved
					
					// In the case of this sample app, this directory itself will be updated properly (moved or removed) if it is a child of another item. That won't happen in one special case: If it is the root item in the list (it has not parent to track it). Fortunately, in our case it doesn't matter if the root directory moved, so we can ignore it. (We'll get a bunch of kAASFileMovedInParentEvent notifications for all of the files in the directory; see below for how we'll handle those.)
					
					break;
				}
				
				if (child == resolvedFile) {
					// Do what needs to be done if this item is an alias and the folder it's pointing to moved.
					
					// In this case, we'd want to keep displaying the items to the user, so we don't have to do anything. (We'll get a bunch of kAASFileMovedInParentEvent notifications for all of the files in the directory; see below for how we'll handle those.)
					
					break;
				}
			
				
				// The file moved out of this directory, it's no longer our concern.
				[self removeChild:child];
				
				break;
				
				
			case kAASFileMovedInParentEvent:
				// We're getting kAASFileMovedInParentEvent notifications for children of this directory because the directory itself moved or some grandparent moved.
				// It is also possible that the file got renamed and the change in the file's path because of the parent/grandparent's movement covered this up.
				
				// We still want to keep the files since they're still within their parent directory. We'll just have to make sure we update to display the appropriate new path information - and this is handled already for us since we're using the file objects themselves for our information storage.
				 // We also have to do whatever needs to be done if this file got renamed - and again in this case this is also handled for us since we're using the file objects for our information storage.
				
				break;
				
			
			case kAASFileRenamedEvent:
				// File in this folder was renamed or the folder itself was renamed
				
				// In our case, renaming will change certain values like the filename, path, and display name. Fortunately, we're using the file objects for our information storage and their information automatically updates during the check - we only need to tell the view to redisplay (which we will do by sending a notification at the end of this update cycle).
				// If we'd been storing additional information someplace other than in the file objects, say for example each SampleItem held an NSAttributedString version of the file name, this would be where we'd update that.

					// Check to make sure it wasn't the directory this item represents that was itself renamed. 
				if (child == file) {
					// Do whatever needs to be done if this folder got renamed
					
					// In this case, nothing
					break;
				}
				
				// Do whatever needs to be done if a file in this folder got renamed
				
				// In this case, nothing
				break;

			case kAASFileRemovedEvent:
				// File in this folder was deleted or the directory itself was deleted
				
					// Check to make sure it wasn't the directory this item represents that was deleted
				if (child == file) {
					// Your folder is gone!  Do something about it!
					
					// In this case, the removal of this directory itself will be updated properly if it is a child of another directory. 
					
					// That won't happen in one special case: If it is the root directory in the list (it has not parent to track it). This is a problem that will have to result in the removal of the directory from the list and its replacement with some default that is known to exist (such as "/").  We'll send a notification to any interested parties and then get the hell out of here - we're about to blink out of existence.
					if (isRoot) {
						[self sendRootRemovalNotification];
						return;
					}
					
					break;
				}

					// If this item is an alias, check to make sure it wasn't the directory that it points to that was deleted
				if (child == resolvedFile) {
					// Your folder is gone!  Do something about it!
					
					// In this case, the alias no longer points to anything.  We have to get rid of its children and then
					// stop treating it like a directory.
					
					[self removeAllChildren];
					
					// Tell our alias file that the item it pointed to is gone
					[file resetAliasStatus]; 
					break;
				}

				// The file was deleted, it's no longer our concern
				[self removeChild:child];
				break;
					
			case kAASFileUpdatedEvent:
				// The file has been updated (data or metadata changed).  Update whatever information you need to update (ie. if you're displaying the file's size, that may have changed).
				
				// In our case, we treat 'updated' events the same as as we do 'renamed' events, for the same reason.
				
				break;
		}
    }
	    
	// Make sure our children are orderly
	[self sortChildren];
	
	// Tell everyone who wants to know that something has changed
	[self sendUpdateNotification];
}


@end
